@@ -1,8 +1,10 @@
 use clap::{Args, Subcommand};
 use eyre::Result;
 
+use atuin_client::record::sqlite_store::SqliteStore;
 use atuin_client::settings::Settings;
 
+pub mod change_password;
 pub mod delete;
 pub mod login;
 pub mod logout;
@@ -27,15 +29,18 @@ pub enum Commands {
 
     // Delete your account, and all synced data
     Delete,
+
+    ChangePassword(change_password::Cmd),
 }
 
 impl Cmd {
-    pub async fn run(self, settings: Settings) -> Result<()> {
+    pub async fn run(self, settings: Settings, store: SqliteStore) -> Result<()> {
         match self.command {
-            Commands::Login(l) => l.run(&settings).await,
+            Commands::Login(l) => l.run(&settings, &store).await,
             Commands::Register(r) => r.run(&settings).await,
             Commands::Logout => logout::run(&settings),
             Commands::Delete => delete::run(&settings).await,
+            Commands::ChangePassword(c) => c.run(&settings).await,
         }
     }
 }
